use std::{
    fmt,
    ops::{Add, AddAssign, Deref, Div, DivAssign, Mul, Sub},
};

mod _2d;
mod shapes;
mod transform;

pub use _2d::*;
pub use shapes::{Hit, Hittable, Shape};
pub use transform::Transform;

#[repr(transparent)]
#[derive(Clone, Copy)]
pub struct Vector(glam::Vec3A);

impl Vector {
    pub const ZERO: Self = Self(glam::Vec3A::ZERO);
    pub const I: Self = Self(glam::Vec3A::X);
    pub const J: Self = Self(glam::Vec3A::Y);
    pub const K: Self = Self(glam::Vec3A::Z);

    pub fn length(self) -> f32 {
        self.0.length()
    }

    pub fn length_squared(self) -> f32 {
        self.0.length_squared()
    }

    pub fn normalize(self) -> Vector {
        Self(self.0.normalize())
    }

    pub fn dot(self, other: Vector) -> f32 {
        self.0.dot(other.0)
    }

    pub fn cross(self, other: Vector) -> Vector {
        Self(self.0.cross(other.0))
    }

    pub fn same_hemisphere(self, other: Vector) -> bool {
        self.dot(other) > 0.0
    }
}

pub fn vec3(x: f32, y: f32, z: f32) -> Vector {
    Vector(glam::vec3a(x, y, z))
}

impl fmt::Debug for Vector {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_tuple(stringify!(Vector))
            .field(&self.0.x)
            .field(&self.0.y)
            .field(&self.0.z)
            .finish()
    }
}

impl From<(f32, f32, f32)> for Vector {
    fn from((x, y, z): (f32, f32, f32)) -> Self {
        vec3(x, y, z)
    }
}

impl From<Point> for Vector {
    fn from(p: Point) -> Self {
        Self(p.0)
    }
}

impl Deref for Vector {
    type Target = glam::Vec3A;
    #[inline]
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<Vector> for glam::Vec3 {
    fn from(value: Vector) -> Self {
        value.0.into()
    }
}

impl Mul<f32> for Vector {
    type Output = Self;

    fn mul(self, rhs: f32) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl Mul<Vector> for f32 {
    type Output = Vector;

    fn mul(self, rhs: Vector) -> Self::Output {
        Vector(self * rhs.0)
    }
}

impl Div<f32> for Vector {
    type Output = Self;

    fn div(self, rhs: f32) -> Self::Output {
        Self(self.0 / rhs)
    }
}

impl DivAssign<f32> for Vector {
    fn div_assign(&mut self, rhs: f32) {
        *self = *self / rhs;
    }
}

impl Add for Vector {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl AddAssign for Vector {
    fn add_assign(&mut self, rhs: Self) {
        *self = *self + rhs;
    }
}

#[repr(transparent)]
#[derive(Clone, Copy)]
pub struct Point(glam::Vec3A);

impl Point {
    pub const ORIGIN: Self = Self(glam::Vec3A::ZERO);
}

pub fn point3(x: f32, y: f32, z: f32) -> Point {
    Point(glam::vec3a(x, y, z))
}

impl fmt::Debug for Point {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_tuple(stringify!(Point))
            .field(&self.0.x)
            .field(&self.0.y)
            .field(&self.0.z)
            .finish()
    }
}

impl From<Point> for glam::Vec3 {
    fn from(value: Point) -> Self {
        value.0.into()
    }
}

impl From<(f32, f32, f32)> for Point {
    fn from((x, y, z): (f32, f32, f32)) -> Self {
        point3(x, y, z)
    }
}

impl Add<Vector> for Point {
    type Output = Point;

    fn add(self, rhs: Vector) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl Sub for Point {
    type Output = Vector;

    fn sub(self, rhs: Self) -> Self::Output {
        Vector(self.0 - rhs.0)
    }
}

impl Sub<Vector> for Point {
    type Output = Point;

    fn sub(self, rhs: Vector) -> Self::Output {
        Point(self.0 - rhs.0)
    }
}

#[derive(Debug)]
pub struct Ray {
    origin: Point,
    direction: Vector,
}

impl Ray {
    pub fn at(&self, t: f32) -> Point {
        self.origin + t * self.direction
    }
}

pub fn ray(origin: Point, direction: Vector) -> Ray {
    Ray { origin, direction }
}
