use std::{
    fmt,
    ops::{Add, AddAssign, Deref, Div, DivAssign, Mul, Sub},
};

#[repr(transparent)]
#[derive(Clone, Copy)]
pub struct Vector2(glam::Vec2);

impl Vector2 {
    pub const ZERO: Self = Self(glam::Vec2::ZERO);
    pub const U: Self = Self(glam::Vec2::X);
    pub const V: Self = Self(glam::Vec2::Y);
}

pub const fn vec2(x: f32, y: f32) -> Vector2 {
    Vector2(glam::vec2(x, y))
}

impl fmt::Debug for Vector2 {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_tuple(stringify!(Vector))
            .field(&self.0.x)
            .field(&self.0.y)
            .finish()
    }
}

impl From<(f32, f32)> for Vector2 {
    fn from((x, y): (f32, f32)) -> Self {
        vec2(x, y)
    }
}

impl Mul<f32> for Vector2 {
    type Output = Self;

    fn mul(self, rhs: f32) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl Mul<Vector2> for f32 {
    type Output = Vector2;

    fn mul(self, rhs: Vector2) -> Self::Output {
        Vector2(self * rhs.0)
    }
}

impl Div<f32> for Vector2 {
    type Output = Self;

    fn div(self, rhs: f32) -> Self::Output {
        Self(self.0 / rhs)
    }
}

impl DivAssign<f32> for Vector2 {
    fn div_assign(&mut self, rhs: f32) {
        *self = *self / rhs;
    }
}

impl Add for Vector2 {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl AddAssign for Vector2 {
    fn add_assign(&mut self, rhs: Self) {
        *self = *self + rhs;
    }
}

#[repr(transparent)]
#[derive(Clone, Copy)]
pub struct Point2(glam::Vec2);

impl Point2 {
    pub const ORIGIN: Self = Self(glam::Vec2::ZERO);
}

pub const fn point2(x: f32, y: f32) -> Point2 {
    Point2(glam::vec2(x, y))
}

impl From<(f32, f32)> for Point2 {
    fn from((x, y): (f32, f32)) -> Self {
        point2(x, y)
    }
}

impl fmt::Debug for Point2 {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_tuple(stringify!(Point))
            .field(&self.0.x)
            .field(&self.0.y)
            .finish()
    }
}

impl Deref for Point2 {
    type Target = glam::Vec2;

    #[inline]
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl Add<Vector2> for Point2 {
    type Output = Point2;

    fn add(self, rhs: Vector2) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl Sub<Vector2> for Point2 {
    type Output = Point2;

    fn sub(self, rhs: Vector2) -> Self::Output {
        Self(self.0 - rhs.0)
    }
}

impl Sub for Point2 {
    type Output = Vector2;

    fn sub(self, rhs: Self) -> Self::Output {
        Vector2(self.0 - rhs.0)
    }
}

#[derive(Copy, Clone)]
#[repr(transparent)]
pub struct Transform2(glam::Affine2);

impl fmt::Debug for Transform2 {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct(stringify!(Transform))
            .field(stringify!(matrix2), &self.0.matrix2)
            .field(stringify!(translation), &self.0.translation)
            .finish()
    }
}

impl Transform2 {
    pub fn scale(x: f32, y: f32) -> Self {
        Self(glam::Affine2::from_scale(glam::vec2(x, y)))
    }

    pub fn inverse(self) -> Self {
        Self(self.0.inverse())
    }

    pub fn translate(x: f32, y: f32) -> Self {
        Self(glam::Affine2::from_translation(glam::vec2(x, y)))
    }
}

impl Mul for Transform2 {
    type Output = Transform2;

    fn mul(self, rhs: Self) -> Self::Output {
        Self(self.0 * rhs.0)
    }
}

impl Mul<Vector2> for Transform2 {
    type Output = Vector2;

    fn mul(self, rhs: Vector2) -> Self::Output {
        Vector2(self.0.transform_vector2(rhs.0))
    }
}

impl Mul<Point2> for Transform2 {
    type Output = Point2;

    fn mul(self, rhs: Point2) -> Self::Output {
        Point2(self.0.transform_point2(rhs.0))
    }
}
